/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::RobotHealth
 * @author     Simon Ottenhaus ( simon dot ottenhaus at kit dot edu )
 * @date       2018
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <array>
#include <atomic>
#include <deque>
#include <mutex>

#include <ArmarXCore/core/Component.h>
#include <ArmarXCore/core/services/tasks/PeriodicTask.h>
#include <ArmarXCore/core/time/DateTime.h>
#include <ArmarXCore/core/time/Duration.h>
#include <ArmarXCore/core/time/Frequency.h>
#include <ArmarXCore/core/time/TimeUtil.h>
#include <ArmarXCore/interface/components/EmergencyStopInterface.h>
#include <ArmarXCore/libraries/ArmarXCoreComponentPlugins/DebugObserverComponentPlugin.h>

#include <ArmarXGui/interface/RemoteGuiInterface.h>

#include <RobotAPI/interface/components/RobotHealthInterface.h>
#include <RobotAPI/interface/speech/SpeechInterface.h>
#include <RobotAPI/interface/units/RobotUnit/RobotUnitInterface.h>

namespace armarx
{
    /**
     * @defgroup Component-RobotHealth RobotHealth
     * @ingroup RobotAPI-Components
     * A description of the component RobotHealth.
     *
     * @class RobotHealth
     * @ingroup Component-RobotHealth
     * @brief Brief description of class RobotHealth.
     *
     * Detailed description of class RobotHealth.
     */
    class RobotHealth :
        virtual public RobotHealthComponentInterface,
        virtual public armarx::Component,
        virtual public armarx::DebugObserverComponentPluginUser
    {
    public:
        /**
         * @see armarx::ManagedIceObject::getDefaultName()
         */
        std::string
        getDefaultName() const override
        {
            return "RobotHealth";
        }

    protected:
        /**
         * @see armarx::ManagedIceObject::onInitComponent()
         */
        void onInitComponent() override;

        /**
         * @see armarx::ManagedIceObject::onConnectComponent()
         */
        void onConnectComponent() override;

        /**
         * @see armarx::ManagedIceObject::onDisconnectComponent()
         */
        void onDisconnectComponent() override;

        /**
         * @see armarx::ManagedIceObject::onExitComponent()
         */
        void onExitComponent() override;

        /**
         * @see PropertyUser::createPropertyDefinitions()
         */
        armarx::PropertyDefinitionsPtr createPropertyDefinitions() override;

    private:
        struct UpdateEntry
        {
            UpdateEntry(const std::string& name,
                        const armarx::core::time::Duration& maximumCycleTimeWarn,
                        const armarx::core::time::Duration& maximumCycleTimeErr) :
                name(name),
                maximumCycleTimeWarn(maximumCycleTimeWarn),
                maximumCycleTimeErr(maximumCycleTimeErr)
            {
            }

            std::string name;
            std::vector<std::string> tags;
            RobotHealthState state = HealthOK;
            std::string description = "";

            bool required = false;
            bool enabled = false;

            mutable std::shared_mutex mutex;

            struct TimeInfo
            {
                //< Timestamp sent by component
                armarx::core::time::DateTime referenceTime;

                //< Timestamp on this PC, set by this component
                armarx::core::time::DateTime arrivalTime; 
            };

            std::deque<TimeInfo> history;

            armarx::core::time::Duration maximumCycleTimeWarn;
            armarx::core::time::Duration maximumCycleTimeErr;
        };

        void monitorHealthUpdateTaskClb();

        UpdateEntry* findUpdateEntry(const std::string& name);
        std::pair<bool, UpdateEntry&> findOrCreateUpdateEntry(const std::string& name);

        void reportDebugObserver();

        // RobotHealthInterface interface
    public:
        void signUp(const RobotHealthHeartbeatArgs& args, const Ice::Current& current) override;
        void unregister(const std::string& identifier, const Ice::Current&) override;
        

        void addRequiredTags(const std::string& requesterIdentifier,
                     const std::vector<std::string>& tags,
                     const Ice::Current& current) override;
        void removeRequiredTags(const std::string& requesterIdentifier,
                       const std::vector<std::string>& tags,
                       const Ice::Current& current) override;
        void resetRequiredTags(const Ice::Current& current) override;


        void heartbeat(const std::string& identifier,
                       const core::time::dto::DateTime& referenceTime,
                       const Ice::Current& current) override;

        std::string getTopicName(const Ice::Current& current) override;

        // RobotHealthComponentInterface interface
    public:
        RobotHealthInfo getSummary(const Ice::Current& current) override;

    private:
        void updateRequiredElements();
        std::set<std::string> requestedTags() const;

        mutable std::shared_mutex updateMutex;

        std::deque<UpdateEntry> updateEntries;

        std::map<std::string, std::set<std::string>> tagsPerRequester;

        PeriodicTask<RobotHealth>::pointer_type monitorUpdateHealthTask;

        armarx::core::time::Duration defaultMaximumCycleTimeWarn;
        armarx::core::time::Duration defaultMaximumCycleTimeErr;

        struct Properties
        {
            EmergencyStopListenerPrx emergencyStopTopicPrx;
            std::string robotHealthTopicName = "RobotHealthTopic";
            AggregatedRobotHealthInterfacePrx aggregatedRobotHealthTopicPrx;
            RemoteGuiInterfacePrx remoteGuiPrx;
            long maximumCycleTimeWarnMS = 50;
            long maximumCycleTimeErrMS = 100;

            std::string requiredTags;
        } p;
    };
} // namespace armarx
