#pragma once

/**
 * Changelog:
 * - Update 1.1.0: Allow templates in aron and anyobjects
 * - Update 1.1.1: Switch to armarx type. Also remove position, orientation and pose types as they are nothing else that matrices or quaternions
 * - Update 2.0.0: Add support for aron defaults. Also add marshalling checks in aron variants.
 **/

#include <ArmarXCore/interface/core/BasicTypes.ice>

// Those macros stay defined
#define ARON_MAJOR "2" // ice interface changes
#define ARON_MINOR "0" // aron header changes
#define ARON_PATCH "0" // aron cpp changes
#define ARON_VERSION "2.0.0"

module armarx
{
    module aron
    {
        // also make version number available to other languages without a preprocessor
        const string MAJOR = ARON_MAJOR;
        const string MINOR = ARON_MINOR;
        const string PATCH = ARON_PATCH;
        const string VERSION = ARON_VERSION;

        /*************************
         * Aron Types ************
         ************************/
        module type
        {
            enum Maybe
            {
                NONE,
                OPTIONAL,
                RAW_PTR,
                UNIQUE_PTR,
                SHARED_PTR
            };

            module ndarray
            {
                enum ElementType
                {
                    UINT8,
                    INT8,
                    UINT16,
                    INT16,
                    UINT32,
                    INT32,
                    FLOAT32,
                    FLOAT64
                };

                module default_value
                {
                    const string ONES = "1";
                    const string ZEROS = "0";
                    const string DEFAULT = "";
                }
            };

            module image
            {
                enum PixelType
                {
                    RGB24,
                    DEPTH32
                };

                module default_value
                {
                    const string IDENTITY = "identity";
                    const string ONES = "1";
                    const string ZEROS = "0";
                    const string DEFAULT = "";
                }
            };

            module pointcloud
            {
                enum VoxelType
                {
                    POINT_XYZ,
                    POINT_XYZI,
                    POINT_XYZL,
                    POINT_XYZRGB,
                    POINT_XYZRGBL,
                    POINT_XYZRGBA,
                    POINT_XYZHSV
                };

                module default_value
                {
                    const string IDENTITY = "identity";
                    const string ONES = "1";
                    const string ZEROS = "0";
                    const string DEFAULT = "";
                }
            };

            module matrix
            {
                enum ElementType
                {
                    INT16,
                    INT32,
                    INT64,
                    FLOAT32,
                    FLOAT64
                };

                module default_value
                {
                    const string IDENTITY = "identity";
                    const string ONES = "1";
                    const string ZEROS = "0";
                    const string DEFAULT = "";
                }
            };

            module quaternion
            {
                enum ElementType
                {
                    FLOAT32,
                    FLOAT64
                };

                module default_value
                {
                    const string ONES = "1";
                    const string ZEROS = "0";
                    const string DEFAULT = "";
                }
            };

            module aron_enum
            {
                module default_value
                {
                    const string DEFAULT = "";
                }
            }

            module dto
            {
                class GenericType
                {
                    string VERSION = ARON_VERSION;
                    Maybe maybe = Maybe::NONE;
                };

                sequence<GenericType> GenericTypeSeq;

                dictionary<string, GenericType> GenericTypeDict;

                /* ***** Container types ***** */
                class List extends GenericType
                {
                    GenericType acceptedType;
                };

                class Tuple extends GenericType

                {
                    GenericTypeSeq elementTypes;
                };

                class Pair extends GenericType
                {
                    GenericType acceptedType1;
                    GenericType acceptedType2;
                };

                class AronObject extends GenericType
                {
                    Ice::StringSeq templates;
                    string objectName;
                    AronObject parent;
                    Ice::StringSeq templateInstantiations;
                    GenericTypeDict elementTypes;
                };

                class Dict extends GenericType
                {
                    GenericType acceptedType;
                }

                /* ***** Complex Types (serialize to ndarray) ***** */
                class NDArray extends GenericType

                {
                    int ndimensions;
                    ndarray::ElementType elementType;
                    string defaultValue = type::ndarray::default_value::DEFAULT;
                    // ONES_DEFAULT_VALUE | ZEROS_DEFAULT_VALUE | <value> | ""
                };

                class Matrix extends GenericType
                {
                    int rows;
                    int cols;
                    matrix::ElementType elementType;
                    string defaultValue = type::matrix::default_value::DEFAULT;
                    // IDENTITY_DEFAULT_VALUE (if quadratic) | ONES_DEFAULT_VALUE | ZEROS_DEFAULT_VALUE | <value> | ""
                };

                class Quaternion extends GenericType
                {
                    quaternion::ElementType elementType;
                    string defaultValue = type::quaternion::default_value::DEFAULT;
                    // ONES_DEFAULT_VALUE | ZEROS_DEFAULT_VALUE | <value sequence> | ""
                };

                class Image extends GenericType
                {
                    image::PixelType pixelType;
                    string defaultValue = type::image::default_value::DEFAULT;
                    // IDENTITY_DEFAULT_VALUE (if quadratic) | ONES_DEFAULT_VALUE | ZEROS_DEFAULT_VALUE | <value> | ""
                };

                class PointCloud extends GenericType
                {
                    pointcloud::VoxelType voxelType;
                    string defaultValue = type::pointcloud::default_value::DEFAULT;
                    // IDENTITY_DEFAULT_VALUE (if quadratic) | ONES_DEFAULT_VALUE | ZEROS_DEFAULT_VALUE | <value> | ""
                };

                /* ***** Enum types ***** */
                class IntEnum extends GenericType
                {
                    string enumName;
                    StringIntDict acceptedValues;
                    string defaultValue = aron_enum::default_value::DEFAULT;
                    // in acceptedValues.keys | ""
                };

                //class FloatEnum extends GenericType { string enumName; StringFloatDict acceptedValues; }
                //class StringEnum extends GenericType { string enumName; StringStringDict acceptedValues; }

                /* ***** Any Types ***** */
                class AnyObject extends GenericType
                {
                };

                /* ***** Primitive Types ***** */
                class AronInt extends GenericType
                {
                    int defaultValue = 0; // <value>
                };

                class AronLong extends GenericType
                {
                    long defaultValue = 0; // <value>
                };

                class AronDouble extends GenericType
                {
                    double defaultValue = 0; // <value>
                };

                class AronFloat extends GenericType
                {
                    float defaultValue = 0; // <value>
                };

                class AronString extends GenericType
                {
                    string defaultValue = ""; // <value>
                };

                class AronBool extends GenericType
                {
                    bool defaultValue = false; // <value>
                };
            };
        };


        /*************************
         * Aron Data *************
         ************************/
        module data
        {
            module dto
            {
                class GenericData
                {
                    string VERSION = ARON_VERSION;
                };

                sequence<GenericData> GenericDataSeq;
                dictionary<string, GenericData> GenericDataDict;

                /* ***** Container Data ***** */
                class List extends GenericData
                {
                    GenericDataSeq elements;
                };

                class Dict extends GenericData
                {
                    GenericDataDict elements;
                };

                /* ***** Complex Data ***** */
                // The NDArray contains more or less the same information as an AronType, but there is no other way to do it
                // Especially, note the difference between the type's typeName (e.g. "GRAY_SCALE_IMAGE" => language dependent) and the data's type ("0")
                // Further, note the difference between the type's dimensions (e.g. 128x128) and the data's dimensions (128x128x3 for RGB)
                class NDArray extends GenericData
                {
                    Ice::IntSeq shape;
                    string type;
                    Ice::ByteSeq data;
                };

                /* ***** Primitive Data ***** */
                // They must be prefixed with 'Aron' as e.g. Int cannot be used as class name
                class AronInt extends GenericData
                {
                    int value;
                };

                class AronLong extends GenericData
                {
                    long value;
                };

                class AronDouble extends GenericData
                {
                    double value;
                };

                class AronFloat extends GenericData
                {
                    float value;
                };

                class AronString extends GenericData
                {
                    string value;
                };

                class AronBool extends GenericData
                {
                    bool value;
                };

                // useful for memory ice_conversions
                sequence<Dict> AronDictSeq;
                sequence<Dict> DictSeq;
            };
        };
    };
};
