#pragma once

#include <SimoxUtility/json.h>

#include <ArmarXCore/core/services/tasks/PeriodicTask.h>

#include <RobotAPI/libraries/armem/core/wm/memory_definitions.h>

namespace armarx::armem::server::ltm::detail::mixin
{
    template <class _CoreSegmentT>
    class BufferedMemoryMixin
    {

    public:
        BufferedMemoryMixin(const MemoryID& id) :
            buffer(std::make_unique<armem::wm::Memory>(id)),
            to_store(std::make_unique<armem::wm::Memory>(id))
        {
        }

        virtual ~BufferedMemoryMixin() = default;

        void
        directlyStore(const armem::wm::Memory& memory)
        {
            std::lock_guard l(storeMutex);

            TIMING_START(LTM_Memory_DirectlyStore);
            _directlyStore(memory);
            TIMING_END_STREAM(LTM_Memory_DirectlyStore, ARMARX_DEBUG);
        }

    protected:
        void
        setMixinMemoryID(const MemoryID& id)
        {
            ARMARX_CHECK_NOT_EMPTY(id.memoryName) << " The full id was: " << id.str();

            buffer->id() = id.getMemoryID();
            to_store->id() = id.getMemoryID();
        }

        void
        start()
        {
            // create task if not already exists
            if (!task)
            {
                int waitingTimeMs = 1000.f / storeFrequency;
                task = new armarx::PeriodicTask<BufferedMemoryMixin>(
                    this, &BufferedMemoryMixin::storeBuffer, waitingTimeMs);
                task->start();
            }
        }

        void
        stop()
        {
            if (task)
            {
                task->stop();
                task = nullptr;
            }
        }

        armem::wm::Memory
        getBuffer() const
        {
            std::lock_guard l(bufferMutex);
            return *buffer;
        }

        void
        storeBuffer()
        {
            std::lock_guard l(storeMutex);
            {
                std::lock_guard l(bufferMutex);
                to_store = std::move(buffer);
                buffer = std::make_unique<armem::wm::Memory>(to_store->id());
            }

            if (to_store->empty())
            {
                ARMARX_INFO << deactivateSpam() << "Cannot store an empty buffer. Ignoring.";
                return;
            }

            this->directlyStore(*to_store);
        }

        /// configuration
        void
        configureMixin(const nlohmann::json& json)
        {
            if (json.find("BufferedMemory.storeFrequency") != json.end())
            {
                storeFrequency = json.at("BufferedMemory.storeFrequency");
            }
        }

        virtual void _directlyStore(const armem::wm::Memory& memory) = 0;

        void
        addToBuffer(const armem::wm::Memory& memory)
        {
            std::lock_guard l(bufferMutex);
            buffer->append(memory);
        }

    protected:
        /// Internal memory for data consolidated from wm to ltm (double-buffer)
        /// The to-put-to-ltm buffer (contains data in plain text)
        /// This buffer may still be filtered (e.g. snapshot filters).
        /// This means that it is not guaranteed that all data in the buffer will be stored in the ltm
        std::unique_ptr<armem::wm::Memory> buffer;
        std::unique_ptr<armem::wm::Memory> to_store;

        /// The frequency (Hz) to store data to the ltm
        float storeFrequency = 10;

    private:
        /// The periodic'task to store the content of the buffer to the ltm
        typename armarx::PeriodicTask<BufferedMemoryMixin>::pointer_type task = nullptr;

        /// a mutex to access the buffer object
        mutable std::recursive_mutex bufferMutex;
        mutable std::recursive_mutex storeMutex;
    };
} // namespace armarx::armem::server::ltm::detail::mixin
