#include "Visu.h"

#include <SimoxUtility/color/Color.h>
#include <SimoxUtility/color/ColorMap.h>
#include <algorithm>
#include <exception>
#include <string>

#include <Eigen/Geometry>

#include <SimoxUtility/algorithm/apply.hpp>
#include <SimoxUtility/algorithm/get_map_keys_values.h>
#include <SimoxUtility/math/pose.h>

#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/time/CycleUtil.h>
#include <ArmarXCore/core/time/TimeUtil.h>
#include <ArmarXCore/interface/core/PackagePath.h>

#include "RobotAPI/components/ArViz/Client/elements/Color.h"
#include "RobotAPI/components/ArViz/Client/elements/PointCloud.h"
#include "RobotAPI/libraries/armem/util/util.h"
#include "RobotAPI/libraries/armem_laser_scans/types.h"
#include "RobotAPI/libraries/core/FramedPose.h"
#include <RobotAPI/components/ArViz/Client/Elements.h>
#include <RobotAPI/interface/components/TrajectoryPlayerInterface.h>
#include <RobotAPI/libraries/armem/core/Time.h>
#include <RobotAPI/libraries/armem_laser_scans/aron/LaserScan.aron.generated.h>
#include <RobotAPI/libraries/armem_laser_scans/aron_conversions.h>
#include <RobotAPI/libraries/armem_laser_scans/util/laser_scanner_conversion.h>

namespace armarx::armem::server::laser_scans
{

    Visu::Visu()
    {
        Logging::setTag("Visu");
    }


    void
    Visu::defineProperties(armarx::PropertyDefinitionsPtr defs, const std::string& prefix)
    {
        defs->optional(
            p.enabled, prefix + "enabled", "Enable or disable visualization of objects.");
        defs->optional(p.frequencyHz, prefix + "frequenzyHz", "Frequency of visualization.");
        defs->optional(p.uniformColor, prefix + "uniformColor", "If enabled, points will be drawn in red.");
        defs->optional(p.maxRobotAgeMs,
                       prefix + "maxRobotAgeMs",
                       "Maximum age of robot state before a new one is retrieved in milliseconds.");
    }


    void
    Visu::init(const wm::CoreSegment* coreSegment, armem::robot_state::VirtualRobotReader* virtualRobotReader)
    {
        this->coreSegment = coreSegment;
        this->virtualRobotReader = virtualRobotReader;
    }


    void
    Visu::connect(const viz::Client& arviz, DebugObserverInterfacePrx debugObserver)
    {
        this->arviz = arviz;
        if (debugObserver)
        {
            bool batchMode = true;
            this->debugObserver = DebugObserverHelper("LaserScansMemory", debugObserver, batchMode);
        }

        if (updateTask)
        {
            updateTask->stop();
            updateTask->join();
            updateTask = nullptr;
        }
        updateTask = new SimpleRunningTask<>([this]() { this->visualizeRun(); });
        updateTask->start();
    }


    void
    Visu::visualizeRun()
    {
        CycleUtil cycle(static_cast<int>(1000 / p.frequencyHz));
        while (updateTask and not updateTask->isStopped())
        {
            if (p.enabled)
            {
                const Time timestamp = Time::Now();
                ARMARX_DEBUG << "Visu task at " << armem::toStringMilliSeconds(timestamp);

                try
                {
                    visualizeOnce(timestamp);
                }
                catch (const std::exception& e)
                {
                    ARMARX_WARNING << "Caught exception while visualizing robots: \n" << e.what();
                }
                catch (...)
                {
                    ARMARX_WARNING << "Caught unknown exception while visualizing robots.";
                }

                if (debugObserver.has_value())
                {
                    debugObserver->sendDebugObserverBatch();
                }
            }
            cycle.waitForCycleDuration();
        }
    }

    void
    Visu::visualizeScan(const std::vector<Eigen::Vector3f>& points,
                        const std::string& sensorName,
                        const std::string& agentName, const viz::Color& color)
    {
        viz::PointCloud pointCloud("laser_scan");

        ARMARX_VERBOSE << "Point cloud with " << points.size() << " points";

        for (const auto& point : points)
        {
            pointCloud.addPoint(point.x(), point.y(), point.z(), color);
        }

        pointCloud.pointSizeInPixels(3);

        viz::Layer l = arviz.layer(agentName + "/" + sensorName);
        l.add(pointCloud);

        arviz.commit(l);
    }

    std::vector<Eigen::Vector3f>
    convertScanToGlobal(const armem::laser_scans::LaserScanStamped& scan,
                        const Eigen::Isometry3f& global_T_sensor)
    {
        auto scanCartesian =
            armarx::armem::laser_scans::util::toCartesian<Eigen::Vector3f>(scan.data);

        for (auto& point : scanCartesian)
        {
            point = global_T_sensor * point;
        }

        return scanCartesian;
    }


    // void Segment::getLatestObjectPoses(const wm::CoreSegment& coreSeg, ObjectPoseMap& out)
    // {
    //     coreSeg.forEachProviderSegment([&out](const wm::ProviderSegment & provSegment)
    //     {
    //         getLatestObjectPoses(provSegment, out);
    //     });
    // }


    // void Segment::getLatestObjectPoses(const wm::ProviderSegment& provSegment, ObjectPoseMap& out)
    // {
    //     provSegment.forEachEntity([&out](const wm::Entity & entity)
    //     {
    //         if (!entity.empty())
    //         {
    //             ObjectPose pose = getLatestObjectPose(entity);
    //             // Try to insert. Fails and returns false if an entry already exists.
    //             const auto [it, success] = out.insert({pose.objectID, pose});
    //             if (!success)
    //             {
    //                 // An entry with that ID already exists. We keep the newest.
    //                 if (it->second.timestamp < pose.timestamp)
    //                 {
    //                     it->second = pose;
    //                 }
    //             }
    //         }
    //     });
    // }


    // void Segment::getLatestObjectPose(const wm::Entity& entity, ObjectPose& out)
    // {
    //     entity.getLatestSnapshot().forEachInstance([&out](const wm::EntityInstance & instance)
    //     {
    //         arondto::ObjectInstance dto;
    //         dto.fromAron(instance.data());

    //         fromAron(dto, out);
    //     });
    // }

    std::map<std::string, armem::laser_scans::LaserScanStamped>
    Visu::getCurrentLaserScans()
    {
        ARMARX_CHECK_NOT_NULL(coreSegment);

        const auto convert = [this](const wm::EntityInstance& entityInstance)
            -> ::armarx::armem::laser_scans::LaserScanStamped
        {
            const std::optional<armarx::armem::laser_scans::arondto::LaserScanStamped> dto =
                tryCast<armarx::armem::laser_scans::arondto::LaserScanStamped>(entityInstance);

            ARMARX_CHECK(dto.has_value());

            ::armarx::armem::laser_scans::LaserScanStamped laserScanStamped;
            fromAron(dto.value(), laserScanStamped);

            const auto ndArrayNavigator =
                aron::data::NDArray::DynamicCast(entityInstance.data()->getElement("scan"));

            ARMARX_CHECK_NOT_NULL(ndArrayNavigator);

            laserScanStamped.data =
                armarx::armem::laser_scans::fromAron<LaserScanStep>(ndArrayNavigator);

            ARMARX_VERBOSE << "Number of steps: " << laserScanStamped.data.size();

            return laserScanStamped;
        };

        std::map<std::string, armem::laser_scans::LaserScanStamped> scans;

        const auto applyToInstance = [&](const wm::EntityInstance& instance)
        {
            const auto scan = convert(instance);
            scans[instance.id().providerSegmentName + "/" + instance.id().entityName] = scan;
        };

        const auto applyToEntity = [&](const wm::Entity& entity)
        {
            if (entity.empty())
            {
                return;
            }

            const auto& snapshot = entity.getLatestSnapshot();

            snapshot.forEachInstance(applyToInstance);
        };

        const auto applyToProviderSegment = [&](const auto& providerSegment)
        { providerSegment.forEachEntity(applyToEntity); };

        coreSegment->forEachProviderSegment(applyToProviderSegment);

        ARMARX_VERBOSE << scans.size() << " scans";
        return scans;
    }

    void
    Visu::visualizeOnce(const Time& timestamp)
    {
        std::map<std::string, armem::laser_scans::LaserScanStamped> currentLaserScans =
            getCurrentLaserScans();

        int i = 0;

        for (const auto& [provider, scan] : currentLaserScans)
        {
            ARMARX_VERBOSE << "Visualizing `" << provider << "`";

            const auto global_T_sensor = [&]() ->  Eigen::Isometry3f{

                const auto robot = getSynchronizedRobot(scan.header.agent, scan.header.timestamp);
                if(not robot)
                {
                    ARMARX_VERBOSE << deactivateSpam(1) << "Robot `" << scan.header.agent << "`" << "not available";
                    return Eigen::Isometry3f::Identity();
                }

                const auto sensorNode = robot->getRobotNode(scan.header.frame);
                ARMARX_CHECK_NOT_NULL(sensorNode) << "No robot node `" << scan.header.frame
                                                << "` for robot `" << scan.header.agent << "`";

                ARMARX_VERBOSE << "Sensor position for sensor `" << scan.header.frame << "` is " << sensorNode->getGlobalPosition();
                return Eigen::Isometry3f{sensorNode->getGlobalPose()};
            }();

            const std::vector<Eigen::Vector3f> points = convertScanToGlobal(scan, global_T_sensor);

            const auto color = [&]() -> simox::Color{ 
                if(p.uniformColor)
                {
                    return simox::Color::red();
                }
                
                return simox::color::GlasbeyLUT::at(i++);
            }();

            visualizeScan(points, scan.header.frame, scan.header.agent, color);
        }
    }

    VirtualRobot::RobotPtr
    Visu::getSynchronizedRobot(const std::string& name, const DateTime& timestamp)
    {
        if (robots.count(name) == 0)
        {
            ARMARX_CHECK_NOT_NULL(virtualRobotReader);
            const auto robot = virtualRobotReader->getRobot(name);

            if (robot)
            {
                robots[name] = {robot, DateTime::Invalid()};
            }
            else
            {
                return nullptr;
            }
        }

        auto& entry = robots.at(name);
        if (entry.second.isInvalid() ||
            (timestamp - entry.second) > Duration::MilliSeconds(p.maxRobotAgeMs))
        {
            if (virtualRobotReader->synchronizeRobot(*entry.first, timestamp))
            {
                entry.second = timestamp;
            }
            else
            {
                ARMARX_INFO << deactivateSpam(10) << "Failed to synchronize robot `" << name << "`";
            }
        }
        return entry.first;
    }

} // namespace armarx::armem::server::laser_scans
