#include "RobotUnitReader.h"

#include <filesystem>
#include <istream>


#include "ArmarXCore/core/time/Frequency.h"
#include "ArmarXCore/core/time/Metronome.h"
#include <ArmarXCore/libraries/ArmarXCoreComponentPlugins/DebugObserverComponentPlugin.h>

#include <RobotAPI/libraries/RobotUnitDataStreamingReceiver/RobotUnitDataStreamingReceiver.h>
#include <RobotAPI/libraries/aron/core/data/variant/container/Dict.h>


namespace armarx::armem::server::robot_state::proprioception
{

    RobotUnitReader::RobotUnitReader() = default;


    void
    RobotUnitReader::connect(armarx::plugins::RobotUnitComponentPlugin& robotUnitPlugin,
                             armarx::plugins::DebugObserverComponentPlugin& debugObserverPlugin,
                             const std::string& robotTypeName)
    {
        {
            converterProprioception = proprioceptionConverterRegistry.get(robotTypeName);
            ARMARX_CHECK_NOT_NULL(converterProprioception)
                << "No converter for robot type '" << robotTypeName << "' available. \n"
                << "Known are: " << proprioceptionConverterRegistry.getKeys();

            config.loggingNames.push_back(properties.sensorPrefix);
            receiver = robotUnitPlugin.startDataStreaming(config);
            description = receiver->getDataDescription();
        }

        {
            converterExteroception = exteroceptionConverterRegistry.get(robotTypeName);
            ARMARX_CHECK_NOT_NULL(converterProprioception)
                << "No converter for robot type '" << robotTypeName << "' available. \n"
                << "Known are: " << exteroceptionConverterRegistry.getKeys();
        }

        {
            // Thread-local copy of debug observer helper.
            debugObserver = DebugObserverHelper(
                Logging::tag.tagName, debugObserverPlugin.getDebugObserver(), true);
        }
    }


    void
    RobotUnitReader::run(float pollFrequency, Queue& dataBuffer)
    {
        Metronome metronome(Frequency::HertzDouble(pollFrequency));

        while (task and not task->isStopped())
        {
            if (std::optional<RobotUnitData> commit = fetchAndConvertLatestRobotUnitData())
            {
                // will lock a mutex
                dataBuffer.push(std::move(commit.value()));
            }

            if (debugObserver)
            {
                debugObserver->sendDebugObserverBatch();
            }

            metronome.waitForNextTick();
        }
    }


    std::optional<RobotUnitData>
    RobotUnitReader::fetchAndConvertLatestRobotUnitData()
    {
        ARMARX_CHECK_NOT_NULL(converterProprioception);

        const std::optional<RobotUnitDataStreaming::TimeStep> data = fetchLatestData();
        if (not data.has_value())
        {
            return std::nullopt;
        }

        ARMARX_DEBUG << "RobotUnitReader: Converting data current timestep to commit";
        auto start = std::chrono::high_resolution_clock::now();

        RobotUnitData result;

        if(converterProprioception != nullptr)
        {
            result.proprioception = converterProprioception->convert(data.value(), description);
        }

        if(converterExteroception != nullptr)
        {
            result.exteroception = converterExteroception->convert(data.value(), description);
        }
        
        result.timestamp = Time(Duration::MicroSeconds(data->timestampUSec));

        auto stop = std::chrono::high_resolution_clock::now();
        auto duration = std::chrono::duration_cast<std::chrono::microseconds>(stop - start);
        ARMARX_DEBUG << "RobotUnitReader: The total time needed to convert the data to commit: "
                     << duration;

        if (debugObserver)
        {
            debugObserver->setDebugObserverDatafield("RobotUnitReader | t Read+Group [ms]",
                                                     duration.count() / 1000.f);
        }

        return result;
    }


    std::optional<RobotUnitDataStreaming::TimeStep>
    RobotUnitReader::fetchLatestData()
    {
        std::deque<RobotUnitDataStreaming::TimeStep>& data = receiver->getDataBuffer();
        if (debugObserver)
        {
            debugObserver->setDebugObserverDatafield("RobotUnitReader | Buffer Size", data.size());
        }
        if (data.empty())
        {
            return std::nullopt;
        }
        else
        {
            RobotUnitDataStreaming::TimeStep currentTimestep = data.back();
            data.clear();
            return currentTimestep;
        }
    }


} // namespace armarx::armem::server::robot_state::proprioception
