/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

// STD/STL
#include <map>
#include <memory>
#include <optional>

// ArmarX
#include <SimoxUtility/xml.h>

#include <ArmarXCore/core/rapidxml/wrapper/RapidXmlReader.h>

/**
 * All constantes for the aron XML parser, in addition to some utility functions wrapping around the armarxcore-xml-parser
 */
namespace armarx::aron::typereader::xml
{
    namespace constantes
    {
        // Definition tags
        const constexpr auto TYPE_DEFINITION_TAG = "arontypedefinition";

        const constexpr auto CODE_INCLUDES_TAG = "codeincludes";
        const constexpr auto ARON_INCLUDES_TAG = "aronincludes";
        const constexpr auto INCLUDES_TAG = "includes";
        const constexpr auto SYSTEM_INCLUDE_TAG = "systeminclude";
        const constexpr auto PACKAGE_PATH_TAG = "packagepath";
        const constexpr auto INCLUDE_TAG = "include";
        const constexpr auto GENERATE_TYPES_TAG = "generatetypes";

        // Attribute names
        const constexpr auto ARON_ATTRIBUTE_NAME = "aron";
        const constexpr auto CODE_ATTRIBUTE_NAME = "code";
        const constexpr auto METHOD_ATTRIBUTE_NAME = "method";
        const constexpr auto RETURN_ATTRIBUTE_NAME = "return";
        const constexpr auto ARGUMENT_TYPE_ATTRIBUTE_NAME = "argumenttype";
        const constexpr auto INCLUDE_ATTRIBUTE_NAME = "include";
        const constexpr auto READER_ATTRIBUTE_NAME = "reader";
        const constexpr auto WRITER_ATTRIBUTE_NAME = "writer";
        const constexpr auto EXTENDS_ATTRIBUTE_NAME = "extends";
        const constexpr auto NAME_ATTRIBUTE_NAME = "name";
        const constexpr auto VALUE_ATTRIBUTE_NAME = "value";
        const constexpr auto TEMPLATE_ATTRIBUTE_NAME = "template";
        const constexpr auto KEY_ATTRIBUTE_NAME = "key";
        const constexpr auto TYPE_ATTRIBUTE_NAME = "type";
        const constexpr auto WIDTH_ATTRIBUTE_NAME = "width";
        const constexpr auto HEIGHT_ATTRIBUTE_NAME = "height";
        const constexpr auto ROWS_ATTRIBUTE_NAME = "rows";
        const constexpr auto COLS_ATTRIBUTE_NAME = "cols";
        const constexpr auto DIMENSIONS_ATTRIBUTE_NAME = "dimensions";
        const constexpr auto SHAPE_ATTRIBUTE_NAME = "shape";
        const constexpr auto OPTIONAL_ATTRIBUTE_NAME = "optional";
        const constexpr auto RAW_PTR_ATTRIBUTE_NAME = "raw_ptr";
        const constexpr auto SHARED_PTR_ATTRIBUTE_NAME = "shared_ptr";
        const constexpr auto UNIQUE_PTR_ATTRIBUTE_NAME = "unique_ptr";
        const constexpr auto PACKAGE_ATTRIBUTE_NAME = "package";
        const constexpr auto DOC_BRIEF_ATTRIBUTE_NAME = "doc-brief";
        const constexpr auto DOC_AUTHOR_ATTRIBUTE_NAME = "doc-author";
        const constexpr auto DOC_PARAM_ATTRIBUTE_NAME = "doc-param";
        const constexpr auto PATH_ATTRIBUTE_NAME = "path";
        const constexpr auto DEFAULT_ATTRIBUTE_NAME = "default";

        // Slugs
        const constexpr auto IDENTITY_SLUG = "identity";
        const constexpr auto ZEROS_SLUG = "zeros";
        const constexpr auto ONES_SLUG = "ones";
        const constexpr auto RGB_SLUG = "rgb24";
        const constexpr auto DEPTH_SLUG = "depth32";
        const constexpr auto POINT_XYZ_SLUG = "pointxyz";
        const constexpr auto POINT_XYZI_SLUG = "pointxyzi";
        const constexpr auto POINT_XYZL_SLUG = "pointxyzl";
        const constexpr auto POINT_XYZRGB_SLUG = "pointxyzrgb";
        const constexpr auto POINT_XYZRGBL_SLUG = "pointxyzrgbl";
        const constexpr auto POINT_XYZRGBA_SLUG = "pointxyzrgba";
        const constexpr auto POINT_XYZHSV_SLUG = "pointxyzhsv";

        const constexpr auto CHAR_SLUG = "int8";
        const constexpr auto UCHAR_SLUG = "uint8";
        const constexpr auto SHORT_SLUG = "int16";
        const constexpr auto USHORT_SLUG = "uint16";
        const constexpr auto INT_SLUG = "int32";
        const constexpr auto UINT_SLUG = "uint32";
        const constexpr auto LONG_SLUG = "int64";
        const constexpr auto ULONG_SLUG = "uint64";
        const constexpr auto FLOAT_SLUG = "float32";
        const constexpr auto DOUBLE_SLUG = "float64";

        // Second level tags. Only important if in specific top level tag
        const constexpr auto OBJECT_CHILD_TAG = "objectchild";
        const constexpr auto ENUM_VALUE_TAG = "enumvalue";

        // Top Level type tags
        const constexpr auto LIST_TAG = "list";
        const constexpr auto DICT_TAG = "dict";
        const constexpr auto OBJECT_TAG = "object";
        const constexpr auto PAIR_TAG = "pair";
        const constexpr auto TUPLE_TAG = "tuple";
        const constexpr auto INT_ENUM_TAG = "intenum";
        const constexpr auto NDARRAY_TAG = "ndarray";
        const constexpr auto MATRIX_TAG = "matrix";
        const constexpr auto QUATERNION_TAG = "quaternion";
        const constexpr auto IMAGE_TAG = "image";
        const constexpr auto POINT_CLOUD_TAG = "pointcloud";
        const constexpr auto CHAR_TAG = "int8";
        const constexpr auto UCHAR_TAG = "uint8";
        const constexpr auto SHORT_TAG = "int16";
        const constexpr auto USHORT_TAG = "uint16";
        const constexpr auto INT_TAG = "int32";
        const constexpr auto UINT_TAG = "uint32";
        const constexpr auto LONG_TAG = "int64";
        const constexpr auto ULONG_TAG = "uint64";
        const constexpr auto FLOAT_TAG = "float32";
        const constexpr auto DOUBLE_TAG = "float64";
        const constexpr auto STRING_TAG = "string";
        const constexpr auto BOOL_TAG = "bool";
        const constexpr auto ANY_OBJECT_TAG = "anyobject";

        // others
        const std::vector<std::string> WHATEVER_VALUES = {"?"};

        struct Replacement
        {
            std::string replacementTag;
            std::pair<std::string, std::string> additionalAronDTOXMLIncludePackagePath;
            std::string deprecatedWarning;
        };

        // Replacements ({tagName, {replacementsTag, additionalAronDTOXMLIncludePackagePath}, deprecationwarning})
        const std::map<std::string, Replacement> REPLACEMENTS = {
            {"int", {"int32", {}, "The <int />-tag is deprecated. Please use <int32 /> instead."}},
            {"long",
             {"int64", {}, "The <long />-tag is deprecated. Please use <int64 /> instead."}},
            {"float",
             {"float32", {}, "The <float />-tag is deprecated. Please use <float32 /> instead."}},
            {"double",
             {"float64", {}, "The <double />-tag is deprecated. Please use <float64 /> instead."}},

            // should that be deprecated?
            {"position", {"matrix rows='3' cols='1' type='float32'", {}, ""}},
            {"pose", {"matrix rows='4' cols='4' type='float32'", {}, ""}},
            {"orientation", {"quaternion type='float32'", {}, ""}},

            // values similar to eigen
            {"quaternionf", {"quaternion type='float32'", {}, ""}},
            {"quaterniond", {"quaternion type='float64'", {}, ""}},
            {"vector2f", {"matrix rows='2' cols='1' type='float32'", {}, ""}},
            {"vector2d", {"matrix rows='2' cols='1' type='float64'", {}, ""}},
            {"vector3f", {"matrix rows='3' cols='1' type='float32'", {}, ""}},
            {"vector3d", {"matrix rows='3' cols='1' type='float64'", {}, ""}},
            {"vector4f", {"matrix rows='4' cols='1' type='float32'", {}, ""}},
            {"vector4d", {"matrix rows='4' cols='1' type='float64'", {}, ""}},
            {"matrix2f", {"matrix rows='2' cols='2' type='float32'", {}, ""}},
            {"matrix2d", {"matrix rows='2' cols='2' type='float64'", {}, ""}},
            {"matrix3f", {"matrix rows='3' cols='3' type='float32'", {}, ""}},
            {"matrix3d", {"matrix rows='3' cols='3' type='float64'", {}, ""}},
            {"matrix4f", {"matrix rows='4' cols='4' type='float32'", {}, ""}},
            {"matrix4d", {"matrix rows='4' cols='4' type='float64'", {}, ""}},

            // You can also add replacements for arondtos here!
            // structure: {xml-identifier, {replacement, auto-include}, deprecationwarning}
            {"datetime",
             {"armarx::arondto::DateTime",
              {"RobotAPI", "libraries/aron/common/aron/time.xml"},
              ""}},
            {"time",
             {"armarx::arondto::DateTime",
              {"RobotAPI", "libraries/aron/common/aron/time.xml"},
              ""}},
            {"duration",
             {"armarx::arondto::Duration",
              {"RobotAPI", "libraries/aron/common/aron/time.xml"},
              ""}},
            {"framedposition",
             {"armarx::arondto::FramedPosition",
              {"RobotAPI", "libraries/aron/common/aron/framed.xml"},
              ""}},
            {"framedorientation",
             {"armarx::arondto::FramedOrientation",
              {"RobotAPI", "libraries/aron/common/aron/framed.xml"},
              ""}},
            {"framedpose",
             {"armarx::arondto::FramedPose",
              {"RobotAPI", "libraries/aron/common/aron/framed.xml"},
              ""}}};
    } // namespace constantes

    namespace util
    {
        bool CheckStrIsInt(const std::string&);
        void EnforceStrIsInt(const std::string&);
        bool CheckStrIsFloat(const std::string&);
        void EnforceStrIsFloat(const std::string&);
        bool CheckStrIsNumber(const std::string&);
        void EnforceStrIsNumber(const std::string&);
        bool CheckStrIsNumberSeq(const std::string&);
        void EnforceStrIsNumberSeq(const std::string&);
        bool CheckStrIsBool(const std::string&);
        void EnforceStrIsBool(const std::string&);

        std::optional<RapidXmlReaderNode> GetFirstNodeWithTag(const RapidXmlReaderNode& node,
                                                              const std::string& name);

        void EnforceAttribute(const RapidXmlReaderNode& node, const std::string& att);
        bool HasAttribute(const RapidXmlReaderNode& node, const std::string& att);
        std::string GetAttribute(const RapidXmlReaderNode& node, const std::string& att);
        std::string GetAttributeWithDefault(const RapidXmlReaderNode& node,
                                            const std::string& att,
                                            const std::string& def);
        bool AttributeIsTrue(const RapidXmlReaderNode& node, const std::string& att);

        bool HasTagName(const RapidXmlReaderNode& node, const std::string& name);
        void EnforceTagName(const RapidXmlReaderNode& node, const std::string& name);
        void EnforceTagNames(const RapidXmlReaderNode& node, const std::vector<std::string>& names);
        std::string GetTagName(const RapidXmlReaderNode& node);

        void EnforceChildSizeSmaller(const RapidXmlReaderNode& node, const size_t size);
        void EnforceChildSizeSmallerEqual(const RapidXmlReaderNode& node, const size_t size);
        void EnforceChildSize(const RapidXmlReaderNode& node, const size_t size);
        void EnforceChildSizeGreaterEqual(const RapidXmlReaderNode& node, const size_t size);
        void EnforceChildSizeGreater(const RapidXmlReaderNode& node, const size_t size);
    } // namespace util
} // namespace armarx::aron::typereader::xml
