#include "SkillProviderComponentPlugin.h"

#include <ArmarXCore/core/Component.h>

#include <RobotAPI/libraries/aron/core/data/variant/primitive/All.h>

namespace armarx::plugins
{
    void
    SkillProviderComponentPlugin::preOnInitComponent()
    {
    }

    void
    SkillProviderComponentPlugin::preOnConnectComponent()
    {
        auto& p = parent<SkillProviderComponentPluginUser>();
        p.getProxy(myPrx, -1);
    }

    void
    SkillProviderComponentPlugin::postOnConnectComponent()
    {
        auto& p = parent<SkillProviderComponentPluginUser>();
        const std::string providerName = p.getName();

        // update skill ownership
        for (auto& [skillName, impl] : skillImplementations)
        {
            impl.skill->manager = manager;
            impl.skill->providerName = providerName;

            impl.statusUpdate.skillId = {providerName, skillName};
        }

        // register self to manager
        skills::manager::dto::ProviderInfo i;
        i.provider = myPrx;
        i.providerName = providerName;
        i.providedSkills = p.getSkillDescriptions();
        manager->addProvider(i);

        connected = true;
    }

    void
    SkillProviderComponentPlugin::preOnDisconnectComponent()
    {
        auto& p = parent<SkillProviderComponentPluginUser>();
        std::string providerName = p.getName();

        manager->removeProvider(providerName);
    }

    void
    SkillProviderComponentPlugin::postCreatePropertyDefinitions(PropertyDefinitionsPtr& properties)
    {
        std::string prefix = "skill.";
        properties->component(
            manager,
            "SkillMemory",
            prefix + "SkillManager",
            "The name of the SkillManager (or SkillMemory) proxy this provider belongs to.");
    }

    void
    SkillProviderComponentPlugin::addSkill(const skills::LambdaSkill::FunT& f,
                                           const skills::SkillDescription& desc)
    {
        auto lambda = std::make_unique<skills::LambdaSkill>(f, desc);
        addSkill(std::move(lambda));
    }

    void
    SkillProviderComponentPlugin::addSkill(std::unique_ptr<skills::Skill>&& skill)
    {
        if (!skill)
        {
            return;
        }

        std::string skillName = skill->description.skillName;
        if (connected) // TODO: fix so that skills can be added anytime!!!
        {
            ARMARX_WARNING << "The SkillProvider already registered to a manager. The skill '" +
                                  skillName +
                                  "' therefore cannot be added anymore. Please only add skills in "
                                  "the onInit method.";
            return;
        }

        // lock skills map
        const std::unique_lock l(skillsMutex);
        if (skillImplementations.find(skillName) != skillImplementations.end())
        {
            ARMARX_WARNING << "Try to add a skill '" + skillName +
                                  "' which already exists in list. Ignoring this skill.";
            return;
        }

        ARMARX_INFO << "Adding skill and set owning provider name" << skillName;
        auto s = skillImplementations.emplace(skillName, std::move(skill));
        s.first->second.statusUpdate.skillId = skills::SkillID(parent().getName(), skillName);
    }

    skills::detail::SkillImplementationWrapper&
    SkillProviderComponentPlugin::getSkill(const std::string& name)
    {
        ARMARX_CHECK_GREATER(skillImplementations.count(name), 0)
            << "Skill '" + name + "' not found.";
        const std::unique_lock l(skillsMutex);
        return skillImplementations.at(name);
    }

    skills::provider::dto::SkillStatusUpdateMap
    SkillProviderComponentPlugin::getSkillExecutionStatuses() const
    {
        skills::provider::dto::SkillStatusUpdateMap skillUpdates;
        const std::unique_lock l(skillsMutex);
        for (const auto& [key, impl] : skillImplementations)
        {
            const std::shared_lock l2(impl.skillStatusMutex);
            skillUpdates.insert({key, impl.statusUpdate.toIce()});
        }
        return skillUpdates;
    }

    skills::provider::dto::SkillDescriptionMap
    SkillProviderComponentPlugin::getSkillDescriptions() const
    {
        skills::provider::dto::SkillDescriptionMap skillDesciptions;
        const std::unique_lock l(skillsMutex);
        for (const auto& [key, impl] : skillImplementations)
        {
            skillDesciptions.insert({key, impl.skill->description.toIce()});
        }
        return skillDesciptions;
    }

} // namespace armarx::plugins

namespace armarx
{
    SkillProviderComponentPluginUser::SkillProviderComponentPluginUser()
    {
        addPlugin(plugin);
    }

    void
    SkillProviderComponentPluginUser::addSkill(std::unique_ptr<skills::Skill>&& skill)
    {
        plugin->addSkill(std::move(skill));
    }

    void
    SkillProviderComponentPluginUser::addSkill(const skills::LambdaSkill::FunT& f,
                                               const skills::SkillDescription& desc)
    {
        plugin->addSkill(f, desc);
    }

    skills::provider::dto::SkillDescription
    SkillProviderComponentPluginUser::getSkillDescription(const std::string& name,
                                                          const Ice::Current& /*unused*/)
    {

        const auto& skillWrapper = plugin->getSkill(name);
        return skillWrapper.skill->description.toIce();
    }

    skills::provider::dto::SkillDescriptionMap
    SkillProviderComponentPluginUser::getSkillDescriptions(const Ice::Current& /*unused*/)
    {
        return plugin->getSkillDescriptions();
    }

    skills::provider::dto::SkillStatusUpdate
    SkillProviderComponentPluginUser::getSkillExecutionStatus(const std::string& skill,
                                                              const Ice::Current& /*unused*/)
    {
        auto& skillWrapper = plugin->getSkill(skill);

        const std::shared_lock l(skillWrapper.skillStatusMutex);
        return skillWrapper.statusUpdate.toIce();
    }

    skills::provider::dto::SkillStatusUpdateMap
    SkillProviderComponentPluginUser::getSkillExecutionStatuses(const Ice::Current& /*unused*/)
    {
        return plugin->getSkillExecutionStatuses();
    }

    // Please not that this method waits until the skill can be scheduled!
    skills::provider::dto::SkillStatusUpdate
    SkillProviderComponentPluginUser::executeSkill(
        const skills::provider::dto::SkillExecutionRequest& info,
        const Ice::Current& /*unused*/)
    {
        // The skill will be executed in a different thread
        std::thread execution;

        // setup input args for skill execution
        skills::SkillParameterization usedParameterization;
        usedParameterization.usedCallbackInterface = info.callbackInterface;
        usedParameterization.usedInputParams = aron::data::Dict::FromAronDictDTO(info.params);

        skills::provider::dto::SkillStatusUpdate ret;
        {
            const std::string skillName = info.skillName;
            auto& wrapper = plugin->getSkill(skillName);

            // async start execution. But we wait for the execution to finish at the end of this method
            execution = std::thread(
                [&ret, &wrapper, &info, &usedParameterization]()
                {
                    // execute waits until the previous execution finishes.
                    auto x = wrapper.setupAndExecuteSkill(info.executorName, usedParameterization);
                    ret = x.toIce();
                });
        } // release lock. We don't know how long the skill needs to finish and we have to release the lock for being able to abort the execution

        if (execution.joinable())
        {
            execution.join();
        }
        return ret;
    }

    void
    SkillProviderComponentPluginUser::abortSkill(const std::string& skillName, const Ice::Current& /*unused*/)
    {
        auto& wrapper = plugin->getSkill(skillName);
        wrapper.skill->notifySkillToStopASAP();
    }

    const std::experimental::observer_ptr<plugins::SkillProviderComponentPlugin>&
    SkillProviderComponentPluginUser::getSkillProviderPlugin() const
    {
        return plugin;
    }
} // namespace armarx
