#pragma once

#include <functional>
#include <queue>
#include <shared_mutex>
#include <thread>
#include <type_traits>
#include <experimental/memory>

#include <ArmarXCore/core/ComponentPlugin.h>
#include <ArmarXCore/core/ManagedIceObject.h>
#include <ArmarXCore/core/services/tasks/RunningTask.h>

#include <RobotAPI/interface/skills/SkillManagerInterface.h>
#include <RobotAPI/libraries/aron/core/data/variant/container/Dict.h>

// Include all types of skills
#include "LambdaSkill.h"
#include "PeriodicSkill.h"
#include "PeriodicSpecializedSkill.h"
#include "Skill.h"
#include "SpecializedSkill.h"

// Helper wrapper for execution
#include "detail/SkillImplementationWrapper.h"
namespace armarx
{
    class SkillProviderComponentPluginUser; // forward declaration
}
namespace armarx::plugins
{
    class SkillProviderComponentPlugin : public ComponentPlugin
    {
    public:
        using ComponentPlugin::ComponentPlugin;

        void preOnInitComponent() override;

        void preOnConnectComponent() override;
        void postOnConnectComponent() override;

        void postCreatePropertyDefinitions(PropertyDefinitionsPtr& properties) override;

        void preOnDisconnectComponent() override;

        void addSkill(const skills::LambdaSkill::FunT&, const skills::SkillDescription&);
        void addSkill(std::unique_ptr<skills::Skill>&&);

        template <typename T>
        T*
        addSkill()
        {
            static_assert(std::is_base_of<skills::Skill, T>::value,
                          "T must be derived from skills::Skill!");
            static_assert(std::is_default_constructible<T>::value,
                          "T must be default constructible!");

            auto skill = std::make_unique<T>();
            auto* skillPtr = skill.get();
            addSkill(std::move(skill));
            return static_cast<T*>(skillPtr);
        }

    private:
        skills::detail::SkillImplementationWrapper& getSkill(const std::string& name);
        skills::provider::dto::SkillStatusUpdateMap getSkillExecutionStatuses() const;
        skills::provider::dto::SkillDescriptionMap getSkillDescriptions() const;

        skills::manager::dti::SkillManagerInterfacePrx manager;
        skills::provider::dti::SkillProviderInterfacePrx myPrx;

        mutable std::shared_mutex skillsMutex;

        bool connected = false;
        std::map<std::string, skills::detail::SkillImplementationWrapper> skillImplementations;

        friend class armarx::SkillProviderComponentPluginUser;
    };
} // namespace armarx::plugins

namespace armarx
{
    class SkillProviderComponentPluginUser :
        virtual public ManagedIceObject,
        virtual public skills::provider::dti::SkillProviderInterface
    {
    public:
        SkillProviderComponentPluginUser();

        skills::provider::dto::SkillDescription
        getSkillDescription(const std::string&,
                            const Ice::Current& current = Ice::Current()) override;
        skills::provider::dto::SkillDescriptionMap
        getSkillDescriptions(const Ice::Current& current = Ice::Current()) override;
        skills::provider::dto::SkillStatusUpdate
        getSkillExecutionStatus(const std::string& skill,
                                const Ice::Current& current = Ice::Current()) override;
        skills::provider::dto::SkillStatusUpdateMap
        getSkillExecutionStatuses(const Ice::Current& current = Ice::Current()) override;

        skills::provider::dto::SkillStatusUpdate
        executeSkill(const skills::provider::dto::SkillExecutionRequest& executionInfo,
                     const Ice::Current& current = Ice::Current()) override;
        void abortSkill(const std::string& name,
                        const Ice::Current& current = Ice::Current()) override;

        const std::experimental::observer_ptr<plugins::SkillProviderComponentPlugin>&
        getSkillProviderPlugin() const;

    protected:
        void addSkill(const skills::LambdaSkill::FunT&, const skills::SkillDescription&);
        void addSkill(std::unique_ptr<skills::Skill>&&);

        template <typename T>
        T*
        addSkill()
        {
            return plugin->addSkill<T>();
        }

    private:
        std::experimental::observer_ptr<plugins::SkillProviderComponentPlugin> plugin = nullptr;
    };
} // namespace armarx
